//------------------------------------------------------------------------
//
// OpenDCC - DCC Sniffer
//
// Copyright (c) 2008 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      tracker.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-04-02 V0.1 kw started; based on the work of Alex Shephard
//                            see www.sourceforge.net, OpenDCC, DCCsniffer
//                            changes made: added individual trackers
//                                          added individual output flags
//                                          put strings to Flash
//            2008-04-17 V0.2 kw added LED control
//            2008-05-15 V0.3 kw added Trigger input
//            2008-05-26 V0.4 kw changed bitfield for FG2 and 3 to 4 Bits
//                               changed text F20-f12 to F20-F13
//
// to do:     accessory PoM commands
//
//------------------------------------------------------------------------
//
// purpose:   DCC sniffer
//            here: evaluate the DCC command and determine the
//            messages to be sent to the host.
//            internal track states are used 
//
// content:   1. States
//            2  Evaluate
//
// uses:      Interface Library (rprintf)
//            LED Conroller (led_control)
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <ctype.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "config.h"
#include "hardware.h"            // get LED defines
#include "led_control.h"         // control the local LEDs
#include "rprintf.h"             // small printf lib from pascal stang
#include "rs485.h"               // structure defines of the DCC Message
#include "TimerPWM.h"            // measure_time
#include "external_trig.h"       // external trigger input

#include "tracker.h"

#define SIMULATION  0            // 0: real application
                                 // 1: test receive routine
                                 // 2: test timing engine
                                 // 3: test action

//=============================================================================
// 1. Definitions
//=============================================================================
//
//
// #define LED_PROG     0     // program   - yellow
// #define LED_RAILCOM  1     // railcom   - yellow
// #define LED_ACC      2     // accessory - green
// #define LED_LOCO     3     // loco      - green
// #define LED_GO       4     // sniffing  - blue
// #define LED_ON       5     // on        - green


t_SnifferState SnifferState;

t_SnifferState eeConfig EEMEM =
  {
    (0<<0) |      // KeyMode: Bit 0: logg only during key pressed
    (1<<1) |      //          Bit 1: key toggles Enable State
    (1<<2),       //          Bit 2: reset timer

    5 |           // DisplayMode -> default Hex5
    D_HEX |       // Display Hex
    D_TSTAMP |    // Display Time Stamp
    D_VERBOSE,    // Display Verbose Messages

    D_ACTIVE |     // TrackTrigger;  
    D_INACTIVE,
    0xff,         // TrackAck;   
    0xff,         // TrackFuture;      // show ServiceMode
    { 0x80,       // TrackInquiry (up to 4 Trackers)
      {3,0,0,0}},    
    { 0x80,       // TrackMessage (up to 4 Trackers) 
      {3,0,0,0}},    
  };



//------------------------------------------------------------------------------
// State Manager

void load_states(void)
  {
	unsigned char i;
    unsigned char  *ramptr;
    
    ramptr = (unsigned char *)&SnifferState;
    
    for (i=0; i < sizeof(SnifferState); i++)
      {
        *ramptr++ = eeprom_read_byte((unsigned char *)&eeConfig+i);
      }
  }	

void save_state_changes(void)
  {
	unsigned char i;
    unsigned char  *ramptr;
    
    ramptr = (unsigned char *)&SnifferState;
    
    for (i=0; i < sizeof(SnifferState); i++)
      {
    	if( *ramptr != eeprom_read_byte((unsigned char *)&eeConfig+i ) )
				eeprom_write_byte( (unsigned char *)&eeConfig+i, *ramptr ) ;
        ramptr++;
      }
  }	



//------------------------------------------------------------------------
// 2. Tracker
//------------------------------------------------------------------------
//

unsigned char should_we_track(t_trackpara *track, unsigned int addr)
  {
    if ( track->bits  & 0x80) return(TRUE);
    if ((track->bits  & (1<<0)) && (track->addr[0] == addr)) return(TRUE);
    if ((track->bits  & (1<<1)) && (track->addr[1] == addr)) return(TRUE);
    if ((track->bits  & (1<<2)) && (track->addr[2] == addr)) return(TRUE);
    if ((track->bits  & (1<<3)) && (track->addr[3] == addr)) return(TRUE);

    return(FALSE);
  }

//------------------------------------------------------------------------
// 2.a) Strings in ProgMemory

char text_stop[]           PROGMEM = "Stop";
char text_estop[]          PROGMEM = "E-Stop";
char text_reserved[]       PROGMEM = "Reserved";
char text_bin_state_ctrl[] PROGMEM = "Bin State Ctrl: [";
char text_dec_ack[]        PROGMEM = "Decoder Acknowledgment";
char text_hard_reset[]     PROGMEM = "Hard Reset";
char text_factory_test[]   PROGMEM = "Factory Test";
char text_dec_flags[]      PROGMEM = "Set Decoder Flags";
char text_adv_addr[]       PROGMEM = "Set Advanced Addressing";
char text_feature_ex[]     PROGMEM = "Feature Expansion, ";
char text_idle[]           PROGMEM = "Idle";

//------------------------------------------------------------------------------
// 2.b) Decoderroutines (generell)
//------------------------------------------------------------------------------
uint8_t displayTimeStamp(void)
  {
    if (SnifferState.DisplayMode & D_TSTAMP) 
      {
        rprintfNum( 10, 4, 0, ' ', measure_time / 1000) ;
        rprintfChar( ',' );
        rprintfNum( 10, 3, 0, '0', measure_time % 1000) ;
        rprintfChar( 's' );
        rprintfChar( ' ' );
        return(TRUE);
      }
    return(FALSE);
  }

// typically we get: header message[0..15] xor {header message[0..15] xor}
uint8_t displayHex(unsigned char* msg, unsigned char msg_size)
  {
    uint8_t i = 0;
    uint8_t temp;
    uint8_t checksum;
    uint8_t Actioned = 0;
    uint8_t local;

    while(i < msg_size)
      {
        local = *msg & 0x0f;              // message size without xor

        rprintfu08(msg[i++]);               // header
        checksum = *msg;
        Actioned = TRUE ;

        for (temp=0; temp < local; temp++)
          {           
            checksum = checksum ^ msg[i];
            // Print a <space> for Hxx1 modes
            if( SnifferState.DisplayMode  & 0x01 ) rprintfChar( ' ' ) ;
            rprintfu08(msg[i++]);
          }

        checksum = checksum ^ msg[i];
        if( SnifferState.DisplayMode  & 0x02 )
          {
            // show xor
            if( SnifferState.DisplayMode  & 0x01 ) rprintfChar( ' ' ) ;
            rprintfu08(msg[i++]);
          }
        else
          {
            i++;
          }
        if (checksum)  rprintfChar( '!' ) ;
      }
    return(Actioned);
  }


// typically we get: header message[0..15] xor {header message[0..15] xor}
void displayChecksum(unsigned char* msg)
  {
    uint8_t temp;
    uint8_t checksum = 0;
    uint8_t local;

    local = *msg & 0x0f;              // message size without xor
    local++;
    for (temp=0; temp <= local; temp++)     // loop covers header, message + xor
      {           
        checksum = checksum ^ msg[temp];
      }

    if (checksum)  rprintfChar( '!' ) ;
    else rprintfChar( ':' ) ;
  }


uint8_t displayTimeAndHex(unsigned char *msg, unsigned char msg_size)
  {
    uint8_t Actioned;

    Actioned = displayTimeStamp();
    Actioned |= displayHex( msg, msg_size );
    return(Actioned);
  }

//------------------------------------------------------------------------------
// 2.c) Decoderroutines (Client messages)
//------------------------------------------------------------------------------

// typically we get: header message[0..15] xor {header message[0..15] xor}
uint8_t displayClient(unsigned char* msg, unsigned char msg_size)
  {
    uint8_t Actioned = 0;
    uint16_t addr;

    if( SnifferState.DisplayMode & D_VERBOSE )
      {
        switch (*msg >> 4)
          {
            case 0x2:
                switch(msg[1])
                  {
                    case 0x10: 
                        // Prog.-Ergebnis anfordern 0x21 0x10 0x31
                        rprintfStr( "PrgRslt?" );
                        Actioned = 1;
                        break;
                    case 0x11: 
                        // Prog.-Lesen Registermode 0x22 0x11 REG X-Or
                        // REG contains the Resister (1...8), this Command has no answer
                        break;
                    case 0x12: 
                        // Prog.-Schreiben Register 0x23 0x12 REG DAT X-Or
                        break;
                    case 0x14:
                        // Prog-lesen Pagemode Hex : 0x22 0x14 CV X-Or-Byte
                        // xp_send_bc_message(PROG_MODE);
                        break;
                    case 0x15: 
                        // Prog.-Lesen CV 0x22 0x15 CV X-Or
                        rprintfStr( "CVrd " );
                        rprintfNum( 10, 3, 0, '0', msg[2]) ;
                        Actioned = 1;
                        break;
                    case 0x16: 
                        // Prog.-Schreiben CV 0x23 0x16 CV DAT X-Or
                        // CV: 1..256
                        rprintfStr( "CVwr " );
                        rprintfNum( 10, 3, 0, '0', msg[2]) ;
                        Actioned = 1;
                        break;
                    case 0x17: 
                        // Prog.-Schreiben Paging 0x23 0x17 CV DAT X-Or
                        break;
                    case 0x21:
                        // Softwareversion anfordern 0x21 0x21 0x00
                        rprintfStr( "Version?" );
                        Actioned = 1;
                        break;
                    case 0x22:
                        // Power Up Mode einstellen 0x22 0x22 00000M00
                        // we don't do this (always manual mode, no automatic power to tracks)
                        break;
                    case 0x24:
                        // Statusabfrage 0x21 0x24 0x05 "Command station status request"
                        rprintfStr( "Status?" );
                        Actioned = 1;
                        break;
                    case 0x80:    
                        // 0x21 0x80 0xA1 "Stop operations request (emergency off)"
                        rprintfStr( "EStop" );
                        Actioned = 1;
                        break;
                    case 0x81:
                        rprintfStr( "Go" );
                        Actioned = 1;
                        // 0x21 0x81 0xA0 "Resume operations request"
                        break;
                  }
                break;
       
            case 0x4:
                // Schaltinformation anfordern 0x42 ADR Nibble X-Or
                // Hex : 0x42 Adresse 0x80 + N X-Or-Byte
                rprintfStr( "FBrd " );
                rprintfu08(msg[1]);
                rprintfu08(msg[2]);
                Actioned = 1;
                break;

            case 0x5:
                // 0x52 Addr DAT [XOR] "Accessory Decoder operation request"
                // Schaltbefehl 0x52 ADR DAT X-Or
                // Hex: 0x52 Adresse 0x80 + SBBO; 
                // Adresse: = Decoder;   S= 1=activate, 0=deactivate,
                //                       BB=local adr,
                //                       O=Ausgang 0 (red) / Ausgang 1 (grn)
                // (das wrde eigentlich schon passend fr DCC vorliegen, aber lieber sauber bergeben)
                addr = (unsigned int) (msg[1] << 2) + ((msg[2] >> 1) & 0b011);
                rprintfStr( "Trnt " );
                rprintfNum( 10, 4, 0, '0', addr) ;
                if (msg[2]  & 0b01) rprintfChar( 'R' );     // coil
                else  rprintfChar( 'G' );
                if (msg[2]  & 0b01000) rprintfChar( '1' );     // activate
                else  rprintfChar( '0' );
                Actioned = 1;
                break;
            case 0x8:
                if (msg[1] == 0x80)          // Alle Loks anhalten 0x80 0x80
                  {
                    rprintfStr( "Halt" );
                    Actioned = 1;
                  }
                break;

            case 0x9:
                // 0x91 loco_addr [XOR] "Emergency stop a locomotive"
                // 0x92 AddrH AddrL [XOR] "Emergency stop a locomotive"
                // 0x9N loco_addr_1 loco_addr_2 etc. loco_addr N [XOR] "Emergency stop selected locomotives"
                if (msg[0] == 2)
                  {
                    addr = ((msg[1] & 0x3F) * 256) + msg[2];
                    rprintfStr( "Stop " );
                    rprintfNum( 10, 4, 0, '0', addr) ;
                    Actioned = 1;
                  }
                break;
            case 0xE:
                switch(msg[1] & 0xf0)      // high nibble von rx_message[1]:
                  {
                    case 0x00:
                        // 0xE3 0x00 AddrH AddrL [XOR] "Locomotive information request"
                        // 0xE4 0x01+R MTR AddrH AddrL [XOR] "Address inquiry member of a Multi-unit request"
                        // 0xE2 0x03+R MTR [XOR] "Address inquiry Multi-unit request"
                        // 0xE3 0x05+R AddrH AddrL [XOR] "Address inquiry locomotive at command station stack request"
                        // 0xE3 0x07 AddrH AddrL [XOR] "Function status request"
                        addr = ((msg[2] & 0x3F) * 256) + msg[3];
                        switch(msg[1] & 0x0f)
                          {
                            case 0x00:
                                rprintfStr( "Lok? " );
                                rprintfNum( 10, 4, 0, '0', addr) ;
                                Actioned = 1;
                                break;
                            case 0x05:
                                rprintfStr( "LokNxt? " );                   // forward scan
                                rprintfNum( 10, 4, 0, '0', addr) ;
                                Actioned = 1;
                                break;
                            case 0x06:
                                rprintfStr( "LokPrv? " );                   // revers scan
                                rprintfNum( 10, 4, 0, '0', addr) ;
                                Actioned = 1;
                                break;
                            case 0x07:
                                rprintfStr( "LokFct? " );
                                rprintfNum( 10, 4, 0, '0', addr) ;
                                Actioned = 1;
                                break;
                          }
                        break;

                    case 0x10:           
                        // Lok Fahrbefehl ab V3 0xE4 Kennung ADR High ADR Low Speed X-Or
                        addr = ((msg[2] & 0x3F) * 256) + msg[3];
                        rprintfStr( "Lok " );
                        rprintfNum( 10, 4, 0, '0', addr) ;
                        rprintfChar( ':');
                        

                        signed char speed;
                        switch(msg[1] & 0x03) // format: 0=14, 1=27, 2=28, 3=128 see t_format Definition
                          {
                            default:
                            case 0:
                                speed = (msg[4] & 0x80) | (msg[4] & 0x0F);
                                break;
                            case 1:
                            case 2:
                                if ((msg[4] & 0x0F) <= 1)               // map 0x?0 to 0 and 0x?1 to 1
                                     speed = msg[4] & 0x81;             // stop or nothalt
                                else 
                                  {
                                    speed = ((msg[4] & 0x0F) << 1) | ((msg[4] & 0x10) >> 4);
                                    speed = speed - 2;                  // map 4..31 to 2..29
                                    speed = speed | (msg[4] & 0x80);    // direction
                                  }
                                break;
                            case 3: // DCC128
                                speed = msg[4];
                                break;
                          }
                        if (speed & 0x80) rprintfChar( 'F');
                        else rprintfChar( 'R');
                        speed &= 0x7F;
                        rprintfNum( 10, 3, 0, '.', speed) ;
                        Actioned = 1;
                        break;
                    case 0x20:
                        addr = ((msg[2] & 0x3F) * 256) + msg[3];
                        switch(msg[1] & 0x0F)   // Lok Funktionsbefehl ab V3 0xE4 Kennung ADR High ADR Low Gruppe X-Or
                          {
                            case 0:          // Hex : 0xE4 0x20 AH AL Gruppe 1 X-Or-Byte   (Gruppe 1: 000FFFFF) f0, f4...f1
                                rprintfStr( "LokF1 " );
                                rprintfNum( 10, 4, 0, '0', addr) ;
                                rprintfChar( ':');
                                rprintfNum(  2, 8, 0, '0', msg[4]) ;
                                Actioned = 1;
                                break;
                            case 1:          // Hex : 0xE4 0x21 AH AL Gruppe 2 X-Or-Byte   (Gruppe 2: 0000FFFF) f8...f5
                                rprintfStr( "LokF2 " );
                                rprintfNum( 10, 4, 0, '0', addr) ;
                                rprintfChar( ':');
                                rprintfNum(  2, 8, 0, '0', msg[4]) ;
                                Actioned = 1;
                                break;

                            case 2:          // Hex : 0xE4 0x22 AH AL Gruppe 3 X-Or-Byte   (Gruppe 3: 0000FFFF) f12...f9
                                rprintfStr( "LokF3 " );
                                rprintfNum( 10, 4, 0, '0', addr) ;
                                rprintfChar( ':');
                                rprintfNum(  2, 8, 0, '0', msg[4]) ;
                                Actioned = 1;
                                break;
                            case 3:          // Hex : 0xE4 0x23 AH AL Gruppe 4 X-Or-Byte   (Gruppe 4: FFFFFFFF) f20...f13
                                rprintfStr( "LokF4 " );
                                rprintfNum( 10, 4, 0, '0', addr) ;
                                rprintfChar( ':');
                                rprintfNum(  2, 8, 0, '0', msg[4]) ;
                                Actioned = 1;
                                break;
                            case 4: // !!! case 5, case 6:
                                            //  Funktionsstatus setzen ab V3 0xE4 Kennung ADR High ADR Low Gruppe X-Or
                                            // Hex : 0xE4 0x24 AH AL Gruppe 1 (000SSSSS)  S=1: Funktion ist tastend
                                            // Hex : 0xE4 0x25 AH AL Gruppe 2 (0000SSSS)
                                            // Hex : 0xE4 0x26 AH AL Gruppe 3 (0000SSSS)
                                break;
                            case 8:          // Hex : 0xE4 0x28 AH AL Gruppe 5 X-Or-Byte   (Gruppe 5: FFFFFFFF) f28...f21
                                break;
                    
                          }
                        break;

                    case 0x30:
                        // Prog. on Main Byte ab V3   0xE6 0x30 AddrH AddrL 0xEC + C CV DAT X-Or
                        // Prog. on Main Bit ab V3    0xE6 0x30 AddrH AddrL 0xE8 + C CV DAT X-Or
                        // Prog. on Main Read ab V3.6 0xE6 0x30 AddrH AddrL 0xEA + C CV DAT [XOR] 

                        // Note: Xressnet does only PoM for Loco, no Accessory!
                        addr = ((msg[2] & 0x3F) * 256) + msg[3];
                          {
                            unsigned int xp_cv;
                            unsigned char xp_data;
                            xp_cv = (msg[4] & 0x03) * 256 + msg[5];
                            xp_data = msg[6];
                            rprintfStr( "PoM " );
                            rprintfNum( 10, 4, 0, '0', addr) ;
                            rprintfChar( ':');
                            rprintfNum( 10, 4, 0, '0', xp_cv) ;
                            rprintfChar( ':');
                            rprintfNum( 10, 3, 0, '0', xp_data) ;
                            Actioned = 1;
                          }
                        break;       

                    case 0x40:   //Lokverwaltung (Double Header)
                        break;
                  }
          }
      }
    return(Actioned);
  }

//------------------------------------------------------------------------------
// 2.d) Decoderroutines (Master messages)
//------------------------------------------------------------------------------

//| - | - | - | 0x61 0x01 0x60 "Normal operation resumed"
//| - | - | - | 0x61 0x00 0x61 "Everything off (emergency off)"
//| - | - | - | 0x81 0x00 0x81 "Everything stopped (emergency stop)"
//| - | - | - | 0x61 0x02 0x63 "Service Mode entry"
//| - | - | - | 0x61 0x12 [XOR] "Programming info: short-circuit"
//| - | - | - | 0x61 0x13 [XOR] "Programming info: Data byte not found"
//| - | - | - | 0x61 0x1f [XOR] "Programming info: Command station busy"
//| - | - | - | 0x61 0x11 [XOR] "Programming info: Command station ready"
//| - | - | - | 0x63 0x10 EE DATA [XOR] "Service Mode response for Register and Paged Mode"
//| - | - | - | 0x63 0x14 CV DATA [XOR] "Service Mode response for Direct CV mode"
//| - | - | - | 0x63 0x15 CV DATA [XOR] "Service Mode response for Direct CV mode (256-511)"
//| - | - | - | 0x63 0x16 CV DATA [XOR] "Service Mode response for Direct CV mode (512-767)"
//| - | - | - | 0x63 0x17 CV DATA [XOR] "Service Mode response for Direct CV mode (768-1023)"                     
//| - | - | - | 0x62 0x21 DATA [XOR] "Software-version"
//| - | - | - | 0x63 0x21 DATA_1 DATA_2 [XOR] "Software-version"
//| - | - | - | 0x62 0x22 DATA [XOR] "Command station status indication"
//| - | - | - | 0x61 0x81 [XOR] "Command station busy"
//| - | - | - | 0x61 0x82 [XOR] "Instruction not supported by command station"
//| - | - | - | 0x61 0x80+F [XOR] "Double Header error"
//| - | - | - | 0x42 Addr DATA [XOR] "Accessory Decoder information"
//| - | - | - | 0x83 Loco_addr Loco_data_1 Loco_data_2 [XOR] "Locomotive is available for operation"
//| - | - | - | 0xA3 Loco_addr Loco_data_1 Loco_data_2 [XOR] "Locomotive is being operated by another device"
//| - | - | - | 0x84 Loco_addr Loco_data_1 Loco_data_2 ModSel [XOR] "Locomotive is available for operation"
//| - | - | - | 0xA4 Loco_addr Loco_data_1 Loco_data_2 ModSel [XOR] "Locomotive is being operated by another device"
//| - | - | - | 0xE4 Identification Speed FKTA FKTB [XOR] "Locomotive information normal locomotive"
//| - | - | - | 0xE5 Identification Speed FKTA FKTB MTR [XOR] "Locomotive information for a locomotive in a multi-unit"
//| - | - | - | 0xE2 Identification Speed [XOR] "Locomotive information for the Multi-unit address"
//| - | - | - | 0xE6 Identification Speed FKTA FKTB AddrH AddrL [XOR] "Locomotive information for a locomotive in a Double Header"
//| - | - | - | 0xE3 0x30+K AddrH AddrL [XOR] "Locomotive information response for address retrieval requests"
//| - | - | - | 0xE3 0x40 AddrH AddrL [XOR] "Locomotive is being operated by another device"
//| - | - | - | 0xE3 0x50 STAT0 STAT1 [XOR] "Function status F0..F12 (momentary)"
//| - | - | - | 0xE3 0x51 STAT0 STAT1 [XOR] "Function status F13..F28 (momentary)"
//| - | - | - | 0xC5 0x04 loco_addr_1 loco_data_1 loco_data_2 loco_addr_2 [XOR] "Double Header available"
//| - | - | - | 0xC5 0x05 loco_addr_1 loco_data_1 loco_data_2 loco_addr_2 [XOR] "Double Header occupied"
//| - | - | - | 0xC6 0x04 loco_addr_1 loco_data_1 loco_data_2 loco_addr_2 ModSel [XOR] "Double Header available"
//| - | - | - | 0xC6 0x05 loco_addr_1 loco_data_1 loco_data_2 loco_addr_2 ModSel [XOR] "Double Header occupied"
//| - | - | - | 0xE1 0x80+F [XOR] "XpressNet error message"

// Lok!F..12/14 00001111 11111111
void displaySpeed(unsigned char speed)
  {
    if (speed == 0) rprintfStr( "..0" );
    else if (speed == 1) rprintfStr( "STP" );
    else
      {
        speed--;
        rprintfNum( 10, 3, 0, '.', speed) ;
      }
  }

void displayLokID(unsigned char *kennung)
  {
    unsigned char speed, temp;

    if (*kennung & 0x08) rprintfChar( '!');         // revers
    else rprintfChar( ':');
    if (kennung[1] & 0x80) rprintfChar( 'F');       // forward
    else rprintfChar( 'R');
    switch(*kennung & 0b0111)
      {
        default:
            rprintfStr( "void" );
            break;
        case 0b000: // DCC 14
            displaySpeed(kennung[1] & 0x7F) ;
            rprintfStr( "/14" );
            break;
        case 0b001:
            speed = kennung[1] & 0x10;
            speed >>= 4;
            temp = kennung[1] & 0x0f;
            temp <<= 1;
            speed = speed | temp;
            displaySpeed(kennung[1] & 0x7F) ;
            rprintfStr( "/27" );
            break;
        case 0b010:
            speed = kennung[1] & 0x10;
            speed >>= 4;
            temp = kennung[1] & 0x0f;
            temp <<= 1;
            speed = speed | temp;
            displaySpeed(kennung[1] & 0x7F) ;
            rprintfStr( "/28" );
            break;
        case 0b100:
            displaySpeed(kennung[1] & 0x7F) ;
            rprintfStr( "/128" );
            break;
      }
  }

uint8_t displayMaster(unsigned char* msg, unsigned char msg_size)
  {
    uint8_t Actioned = 0;
    uint16_t addr;

    if( SnifferState.DisplayMode & D_VERBOSE )
      {
        switch (*msg >> 4)
          {
            case 0x4:
                // 0x42 Addr DATA [XOR] "Accessory Decoder information"
                break;
            case 0x6:
                switch (msg[1])
                  {
                    case 0x00:
                        // 0x61 0x00 0x61 "Everything off (emergency off)"
                        rprintfStr( "Off" );
                        Actioned = 1;
                        break;
                    case 0x01:
                        // 0x61 0x01 0x60 "Normal operation resumed"
                        rprintfStr( "Go" );
                        Actioned = 1;
                        break;
                    case 0x02:
                        // 0x61 0x02 0x63 "Service Mode entry"
                        rprintfStr( "SM-entry" );
                        Actioned = 1;
                        break;
                        // 0x63 0x10 EE DATA [XOR] "Service Mode response for Register and Paged Mode"
                        // 0x61 0x11 [XOR] "Programming info: Command station ready"
                        // 0x61 0x12 [XOR] "Programming info: short-circuit"
                        // 0x61 0x13 [XOR] "Programming info: Data byte not found"
                        // 0x63 0x14 CV DATA [XOR] "Service Mode response for Direct CV mode"
                        // 0x63 0x15 CV DATA [XOR] "Service Mode response for Direct CV mode (256-511)"
                        // 0x63 0x16 CV DATA [XOR] "Service Mode response for Direct CV mode (512-767)"
                        // 0x63 0x17 CV DATA [XOR] "Service Mode response for Direct CV mode (768-1023)"
                        // 0x61 0x1f [XOR] "Programming info: Command station busy"
                    case 0x21:
                        // 0x62 0x21 DATA [XOR] "Software-version"
                        // 0x63 0x21 DATA_1 DATA_2 [XOR] "Software-version"
                        rprintfStr( "Version " );
                        rprintfu04(msg[2]>>4);
                        rprintfChar( '.');
	                    rprintfu04(msg[2]);
                        if ((*msg & 0xF) == 3) {rprintfChar( ':'); rprintfu08(msg[3]);}
                        Actioned = 1;
                        break;
                    case 0x22:
                        // 0x62 0x22 DATA [XOR] "Command station status indication"
                        rprintfStr( "Status " );
                        if ((msg[2] & 0x03) == 0) rprintfStr( "++" );
                        else
                          {
                            if (msg[2] & 0x01) rprintfChar( 'H');   // Halt
                            else rprintfChar( '.');
                            if (msg[2] & 0x02) rprintfChar( 'S');   // Stop
                            else rprintfChar( '.');
                          }
                        if (msg[2] & 0x04) rprintfChar( 'a');
                        else rprintfChar( 'm');
                        if (msg[2] & 0x08) rprintfChar( 'P');       // Progmode
                        else rprintfChar( '.');
                        rprintfStr( "__" );
                        if (msg[2] & 0x40) rprintfChar( 'B');
                        else rprintfChar( '_');
                        if (msg[2] & 0x80) rprintfChar( 'E');
                        else rprintfChar( '_');
                        Actioned = 1;
                        break;
                    case 0x81:
                        // 0x61 0x81 [XOR] "Command station busy"
                        rprintfStr( "busy" );
                        Actioned = 1;
                        break;
                    case 0x82:
                        // 0x61 0x82 [XOR] "Instruction not supported by command station"
                        rprintfStr( "not supp." );
                        Actioned = 1;
                        break;                        
                        // 0x61 0x80+F [XOR] "Double Header error"
                  }
                break;
            case 0x8:
                // 0x81 0x00 0x81 "Everything stopped (emergency stop)"
                // 0x83 Loco_addr Loco_data_1 Loco_data_2 [XOR] "Locomotive is available for operation"
                // 0x84 Loco_addr Loco_data_1 Loco_data_2 ModSel [XOR] "Locomotive is available for operation"
                break;
            case 0xA:
                // 0xA3 Loco_addr Loco_data_1 Loco_data_2 [XOR] "Locomotive is being operated by another device"
                // 0xA4 Loco_addr Loco_data_1 Loco_data_2 ModSel [XOR] "Locomotive is being operated by another device"
                break;
            case 0xC:
                // 0xC5 0x04 loco_addr_1 loco_data_1 loco_data_2 loco_addr_2 [XOR] "Double Header available"
                // 0xC5 0x05 loco_addr_1 loco_data_1 loco_data_2 loco_addr_2 [XOR] "Double Header occupied"
                // 0xC6 0x04 loco_addr_1 loco_data_1 loco_data_2 loco_addr_2 ModSel [XOR] "Double Header available"
                // 0xC6 0x05 loco_addr_1 loco_data_1 loco_data_2 loco_addr_2 ModSel [XOR] "Double Header occupied"
                break;
            case 0xE:
                switch(msg[1] & 0xF0)
                  {
                    case 0x00:
                        // 0xE4 Identification Speed FKTA FKTB [XOR] "Locomotive information normal locomotive"
                        rprintfStr( "Lok" );
                        displayLokID(&msg[1]);
                        rprintfChar( '.');
                        rprintfNum( 2, 8, 0, '0', msg[3]) ;                
                        rprintfChar( '.');
                        rprintfNum( 2, 8, 0, '0', msg[4]) ;                
                        Actioned = 1;
                        break;
                        // 0xE5 Identification Speed FKTA FKTB MTR [XOR] "Locomotive information for a locomotive in a multi-unit"
                        // 0xE2 Identification Speed [XOR] "Locomotive information for the Multi-unit address"
                        // 0xE6 Identification Speed FKTA FKTB AddrH AddrL [XOR] "Locomotive information for a locomotive in a Double Header"
                        // 0xE3 0x30+K AddrH AddrL [XOR] "Locomotive information response for address retrieval requests"
                    case 0x40:
                        // 0xE3 0x40 AddrH AddrL [XOR] "Locomotive is being operated by another device"
                        rprintfStr( "Lok " );
                        addr = (msg[2] & 0x3F) * 256 + msg[3];
                        rprintfNum( 10, 4, 0, '0', addr) ;
                        rprintfStr( " stolen" );
                        Actioned = 1;
                        break;
                        // 0xE3 0x50 STAT0 STAT1 [XOR] "Function status F0..F12 (tastend)"
                        // 0xE3 0x51 STAT0 STAT1 [XOR] "Function status F13--F28 (tastend)"
                        // 0xE1 0x80+F [XOR] "XpressNet error message"
                  }
                break;
          }
      }
    return(Actioned);
  }

//-----------------------------------------------------------------------
// xpressnet
//-----------------------------------------------------------------------
// 
// * msg = pointer to message (callbyte)
// msg_size = nr. of chr, including xor

void displayXPmsg(unsigned char* msg, unsigned char msg_size)
  {
    unsigned char Actioned = 0; // true if any decoding happened
    unsigned char Verbosed = 0; // true if verbose mode decoding happened
    
    // first, check callbyte

    unsigned char slot, temp;

    
    temp = *msg ^ (*msg >> 4);
    temp = temp ^ (temp >> 2);
    temp = temp ^ (temp >> 1);
    if (temp & 0x01)
      {
        if (SnifferState.DisplayMode & D_VERBOSE)
              {
                rprintfStr( "CallPErr" );
                Actioned = 1 ;
              } 
      }
    slot = *msg & 0x1F;
    switch(*msg++ & 0x60)
      {
        case 0x00:      // ACK
            if( SnifferState.TrackAck)
              {
                displayTimeStamp();
                rprintfChar( 'A' );
                rprintfNum( 10, 2, 0, '0', slot) ;                
                if (msg_size > 0) 
                  {
                    displayChecksum(msg);
                    displayHex(msg, msg_size-1);
                  } 
                Actioned = 1;
              }
            break;
        case 0x20:      // Future or Feedback
            if( SnifferState.TrackFuture)
              {
                displayTimeStamp();
                if (SnifferState.DisplayMode & D_VERBOSE)
                  {
                    if (slot == 0 && (*msg++ & 0xf0) == 0x40 && msg_size >=5)
                      {
                        rprintfChar( 'F' );rprintfChar( 'B' );rprintfChar( ' ' );
                        displayChecksum(msg);
                        unsigned char j;
                        for (j=0; j<msg_size-3; j+=2)
                          {
                            rprintfNum( 10, 4, 0, '0', msg[j] * 8 + 4 * ((msg[j+1] & 0x10) >> 4)    ) ;
                            rprintfChar( ':' );
                            rprintfNum( 2, 4, 0, '0', msg[j+1] & 0x0F) ;
                          }
                        Actioned = 1 ;
                        Verbosed = 1 ;
                      }
                    if (!Verbosed)
                      {
                        rprintfChar( 'F' );
                        rprintfNum( 10, 2, 0, '0', slot) ;
                        displayChecksum(msg);
                        if (msg_size > 0) displayHex(msg, msg_size-1);
                        Actioned = 1;
                      }
                  }
                else
                  {
                    rprintfChar( 'F' );
                    rprintfNum( 10, 2, 0, '0', slot) ;
                    displayChecksum(msg);
                    if (SnifferState.DisplayMode & D_HEX) 
                      {
                        if (msg_size > 0) displayHex(msg, msg_size-1);
                      }
                    Actioned = 1;
                  }
              }
            break;
        case 0x40:      // Inquiry
            if (msg_size > 1)
              {  // Inquiry with following request from client
                if (should_we_track (&SnifferState.TrackMessage, slot) )
                  {          
                    Actioned |= displayTimeStamp();
                    rprintfChar( 'C' );
                    rprintfNum( 10, 2, 0, '0', slot) ;
                    displayChecksum(msg);

                    if( SnifferState.DisplayMode & D_HEX )
                      {
                        displayHex(msg, msg_size-1);
                      }
                    if (SnifferState.DisplayMode & D_VERBOSE) 
                      {
                        Verbosed = displayClient(msg, msg_size-1); 
                        if (!Verbosed) displayHex(msg, msg_size-1);
                      }                    
                    Actioned = 1 ;
                  }
              }
            else
              {
                if (should_we_track (&SnifferState.TrackInquiry, slot) )
                  {          
                    Actioned |= displayTimeStamp();
                    rprintfChar( 'I' );
                    rprintfNum( 10, 2, 0, '0', slot) ;
                    Actioned = 1 ;
                  }
              }
            break;
        case 0x60:      // master message
            if (should_we_track (&SnifferState.TrackMessage, slot) )
              {          
                Actioned |= displayTimeStamp();
                rprintfChar( 'M' );
                rprintfNum( 10, 2, 0, '0', slot) ;
                displayChecksum(msg);

                if( SnifferState.DisplayMode & D_HEX )
                  {
                    displayHex(msg, msg_size-1);
                  }
                if (SnifferState.DisplayMode & D_VERBOSE) 
                  {
                    Verbosed = displayMaster(msg, msg_size-1); 
                    if (!Verbosed) displayHex(msg, msg_size-1);
                  }                    
                Actioned = 1 ;
              }
            break;
      }

    if (Actioned)    rprintfCRLF();
  }


//-----------------------------------------------------------------------
// external Trigger
//-----------------------------------------------------------------------
// watch the trigger input and create a message

void watch_external_trigger(void)
  {
    unsigned char state;

    state = trigger_poll();

    if (state == 0)
      {
        if (SnifferState.TrackTrigger & D_INACTIVE)
          {
            displayTimeStamp();
            rprintfStr( "Trig inactive" );
            rprintfCRLF();        
          }
      }
    else if (state == 1)
      {
        if (SnifferState.TrackTrigger & D_ACTIVE)
          {
            displayTimeStamp();
            rprintfStr( "Trig active" );
            rprintfCRLF();        
          }
      }
  }






//------------------------------------------------------------------------
// 
// 3. Upstream Interface
//
//------------------------------------------------------------------------

unsigned char EnableDisplay;

void toggle_EnableDisplay(void)
  {
    if (EnableDisplay)
      {
        EnableDisplay = FALSE;
        led_off(LED_GO);
      }
    else
      {
        EnableDisplay = TRUE;
        led_half(LED_GO);  
      }
  }


unsigned char xmsg[64];
unsigned char xmsg_index;

int8_t last_msg;    // signed

void init_tracker(void)
  {
    xmsg_index = 0;
    load_states();              // get all states from eeprom
    EnableDisplay = FALSE;
    toggle_EnableDisplay();
    last_msg = systick_1ms;     // keep track of last dcc msg
  }


char text_overrun[]   PROGMEM = "Overrun!";


void run_tracker(void)
  {
    watch_external_trigger();

    if (XP_rx_ready())
      {
        unsigned int xdata;

        xdata = XP_rx_read();                 // reads one byte

        if (xdata >= 256)
          { // new call byte
            // process current slot
            if ( EnableDisplay) displayXPmsg(xmsg, xmsg_index);
            
            xmsg_index = 0;
          }
        xmsg[xmsg_index] = (unsigned char) xdata;
        xmsg_index++;
        if (xmsg_index >= 64)
          {
            if ( EnableDisplay)
              {
                rprintfProgStr( text_overrun ) ;  
                rprintfCRLF() ;
              }
          }
      }
  }


/*
void run_spy(void)
  {
    if (XP_rx_ready())
      {
        unsigned int xdata;

        LED_RAILCOM_ON;

        xdata = XP_rx_read();                 // reads one byte
        tx_fifo_write((unsigned char) xdata);
        LED_RAILCOM_OFF;
      }
  }
*/
